# Copyright 2023 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""bundle_refactor

Revision ID: f7843ddf1552
Revises: 151893067f91
Create Date: 2023-09-16 09:25:00.674820

"""

# revision identifiers, used by Alembic.
revision = 'f7843ddf1552'
down_revision = '151893067f91'
branch_labels = None
depends_on = None

import logging

from alembic import op
import sqlalchemy as sa


REF_TABLE = 'zuul_ref'
BUILD_TABLE = 'zuul_build'
BUILDSET_TABLE = 'zuul_buildset'
BUILDSET_REF_TABLE = 'zuul_buildset_ref'
ARTIFACT_TABLE = 'zuul_artifact'
BUILD_EVENT_TABLE = 'zuul_build_event'
PROVIDES_TABLE = 'zuul_provides'


def rename_index(connection, table, old, new):
    dialect_name = connection.engine.dialect.name
    if dialect_name in ('mysql', 'mariadb'):
        statement = f"""
            alter table {table}
            rename index {old}
            to {new}
        """
    elif dialect_name == 'postgresql':
        statement = f"""
            alter index {old}
            rename to {new}
        """
    else:
        raise Exception(f"Unsupported dialect {dialect_name}")
    connection.execute(sa.text(statement))


# This migration has an unusual structure.  We do quite a bit of work
# on temporary tables before we start replacing our real tables.
# Mysql doesn't support transactional DDL, but we can fairly easily
# undo our changes up to a certain point just by dropping our
# temporary tables.  To make life easier on operators in case there is
# a problem, if we encounter an error up to the point of no return, we
# will drop the tables and leave the db in the state it started.  If
# we encounter an error after that point, there is little we can do to
# recover, but hopefully we will have found errors prior to then, if
# any.

# Postgres has transactional DDL and the entire migraction is in a
# single transaction and is automatically rolled back.  We just no-op
# in that case.

# To accomodate this structure without having a giant try/except
# block, the migration is split into 2 functions, upgrade1() and
# upgrade2().

def upgrade1(connection, table_prefix):
    # This is the first part of the migration, which is recoverable in
    # mysql.
    prefixed_ref = table_prefix + REF_TABLE
    prefixed_build = table_prefix + BUILD_TABLE
    prefixed_build_new = table_prefix + BUILD_TABLE + '_new'
    prefixed_buildset = table_prefix + BUILDSET_TABLE
    prefixed_buildset_new = table_prefix + BUILDSET_TABLE + '_new'
    prefixed_buildset_ref = table_prefix + BUILDSET_REF_TABLE
    quote = connection.engine.dialect.identifier_preparer.quote
    dialect_name = connection.engine.dialect.name

    # This migration requires updates to existing rows in the
    # zuul_build and zuul_buildset tables.  In postgres, tables have a
    # fill factor which indicates how much space to leave in pages for
    # row updates.  With a high fill factor (the default is 100%)
    # large updates can be slow.  With a smaller fill factor, large
    # updates can bu much faster, at the cost of wasted space and
    # operational overhead.  The default of 100% makes sense for all
    # of our tables.  While the build and buildset tables do get some
    # row updates, they are not very frequent.  We would need a very
    # generous fill factor to be able to update all of the rows in the
    # build table quickly, and that wouldn't make sense for normal
    # operation.

    # Instead of adding columns and updating the table, we will
    # create new tables and populate them with inserts (which is
    # extremely fast), then remove the old tables and rename the new.

    # First, create zuul_buildset_new table.  This will later replace
    # the zuul_buildset table.  It includes some changes:

    # * We intentionally omit the zuul_ref column (in this case,
    #   referring to the old Z<sha> unique ids) because it is obsolete
    #   is obsolete.

    # * The length of the sha fields is lowered to 40.  This is so
    #   that in the zuul_ref table, we can make a unique index of
    #   several fields without hitting mysql's index length limit.
    #   We will mutate their values as we insert them here.
    op.create_table(
        prefixed_buildset_new,
        sa.Column('id', sa.Integer, primary_key=True),
        sa.Column('pipeline', sa.String(255)),
        sa.Column('message', sa.TEXT()),
        sa.Column('tenant', sa.String(255)),
        sa.Column('result', sa.String(255)),
        sa.Column('uuid', sa.String(36)),
        sa.Column('event_id', sa.String(255), nullable=True),
        sa.Column('event_timestamp', sa.DateTime, nullable=True),
        sa.Column('first_build_start_time', sa.DateTime, nullable=True),
        sa.Column('last_build_end_time', sa.DateTime, nullable=True),
        sa.Column('updated', sa.DateTime, nullable=True),
        # Columns we'll drop later, but need now in order to populate
        # zuul_ref.
        sa.Column('project', sa.String(255), nullable=False),
        sa.Column('ref', sa.String(255), nullable=False),
        sa.Column('ref_url', sa.String(255), nullable=False),
        sa.Column('change', sa.Integer, nullable=False),
        sa.Column('patchset', sa.String(40), nullable=False),
        sa.Column('oldrev', sa.String(40), nullable=False),
        sa.Column('newrev', sa.String(40), nullable=False),
        sa.Column('branch', sa.String(255), nullable=False),
    )

    # The postgres operator "is not distinct from" (equivalent to
    # mysql's <=>) is a non-indexable operator.  So that we can
    # actually use the unique index (and other indexes in the future)
    # make all of the ref-related columns non-null.  That means empty
    # strings for strings, and we'll use 0 for the change id.  This
    # lets us use the "=" operator and utilize the index for all
    # values.
    statement = f"""
        insert into {prefixed_buildset_new}
            select bs.id, bs.pipeline, bs.message,
                bs.tenant, bs.result, bs.uuid, bs.event_id, bs.event_timestamp,
                bs.first_build_start_time, bs.last_build_end_time, bs.updated,
                bs.project, coalesce(bs.ref, ''), coalesce(ref_url, ''),
                coalesce(bs.change, 0), coalesce(bs.patchset, ''),
                coalesce(bs.oldrev, ''), coalesce(bs.newrev, ''),
                coalesce(branch, '')
            from {prefixed_buildset} bs
    """
    connection.execute(sa.text(statement))

    # Create zuul_ref table.
    op.create_table(
        prefixed_ref,
        sa.Column('id', sa.Integer, primary_key=True),
        sa.Column('project', sa.String(255), nullable=False),
        sa.Column('ref', sa.String(255), nullable=False),
        sa.Column('ref_url', sa.String(255), nullable=False),
        sa.Column('change', sa.Integer, nullable=False),
        sa.Column('patchset', sa.String(40), nullable=False),
        sa.Column('oldrev', sa.String(40), nullable=False),
        sa.Column('newrev', sa.String(40), nullable=False),
        sa.Column('branch', sa.String(255), nullable=False),
    )

    # Copy data from buildset to ref.

    # We are going to have a unique index later on some columns, so we
    # use a "group by" clause here to remove duplicates.  We also may
    # have differing values for ref_url for the same refs (e.g.,
    # opendev switched gerrit server hostnames), so we arbitrarily
    # take the first ref_url for a given grouping.  It doesn't make
    # sense for branch to be different, but we do the same in order to
    # avoid any potential errors.
    statement = f"""
        insert into {prefixed_ref}
            (project, {quote('change')}, patchset,
             ref, ref_url, oldrev, newrev, branch)
        select
            bs.project, bs.change, bs.patchset,
            bs.ref, min(ref_url), bs.oldrev, bs.newrev, min(branch)
        from {prefixed_buildset_new} bs
        group by
            bs.project, bs.change, bs.patchset,
            bs.ref, bs.oldrev, bs.newrev
    """
    connection.execute(sa.text(statement))

    # Create our unique ref constraint; this includes as index that
    # will speed up populating the zuul_buildset_ref table.
    op.create_unique_constraint(
        f'{prefixed_ref}_unique',
        prefixed_ref,
        ['project', 'ref', 'change', 'patchset', 'oldrev', 'newrev'],
    )

    # Create replacement indexes for the obsolete indexes on the
    # buildset table.
    with op.batch_alter_table(prefixed_ref) as batch_op:
        batch_op.create_index(
            f'{prefixed_ref}_project_change_idx',
            ['project', 'change'])
        batch_op.create_index(
            f'{prefixed_ref}_change_idx',
            ['change'])

    # Add mapping table for buildset <-> ref

    # We will add foreign key constraints after populating the table.
    op.create_table(
        prefixed_buildset_ref,
        sa.Column('buildset_id', sa.Integer, nullable=False),
        sa.Column('ref_id', sa.Integer, nullable=False),
        sa.PrimaryKeyConstraint("buildset_id", "ref_id"),
    )

    # Populate buildset_ref table.  Ignore ref_url since we don't
    # include it in the unique index later.
    statement = f"""
        insert into {prefixed_buildset_ref}
        select {prefixed_buildset_new}.id, {prefixed_ref}.id
        from {prefixed_buildset_new} left join {prefixed_ref}
        on {prefixed_buildset_new}.project = {prefixed_ref}.project
        and {prefixed_buildset_new}.ref = {prefixed_ref}.ref
        and {prefixed_buildset_new}.change = {prefixed_ref}.change
        and {prefixed_buildset_new}.patchset = {prefixed_ref}.patchset
        and {prefixed_buildset_new}.oldrev = {prefixed_ref}.oldrev
        and {prefixed_buildset_new}.newrev = {prefixed_ref}.newrev
    """
    connection.execute(sa.text(statement))

    # Fix the sequence value since we wrote our own ids (postgres only)
    if dialect_name == 'postgresql':
        statement = f"""
            select setval(
                 '{prefixed_buildset_new}_id_seq',
                 COALESCE((SELECT MAX(id)+1 FROM {prefixed_buildset_new}), 1),
                 false)
        """
        connection.execute(sa.text(statement))

    # Now that the table is populated, add the FK indexes and
    # constraints to buildset_ref.
    op.create_index(
        f'{prefixed_buildset_ref}_buildset_id_idx',
        prefixed_buildset_ref, ['buildset_id'])
    op.create_index(
        f'{prefixed_buildset_ref}_ref_id_idx',
        prefixed_buildset_ref, ['ref_id'])

    # Alembic doesn't allow us to combine alter table operations, so
    # we do this manually.  It still takes the same total time in
    # mysql though.
    statement = f"""
        alter table {prefixed_buildset_ref}
        add constraint {prefixed_buildset_ref}_buildset_id_fkey
            foreign key(buildset_id)
            references {prefixed_buildset_new} (id),
        add constraint {prefixed_buildset_ref}_ref_id_fkey
            foreign key(ref_id)
            references {prefixed_ref} (id)
    """
    connection.execute(sa.text(statement))

    # Our goal below is to add the ref_id column to the build table
    # and populate it with a query.

    # Create the new build table.
    op.create_table(
        prefixed_build_new,
        sa.Column('id', sa.Integer, primary_key=True),
        sa.Column('buildset_id', sa.Integer),
        sa.Column('uuid', sa.String(36)),
        sa.Column('job_name', sa.String(255)),
        sa.Column('result', sa.String(255)),
        sa.Column('start_time', sa.DateTime),
        sa.Column('end_time', sa.DateTime),
        sa.Column('voting', sa.Boolean),
        sa.Column('log_url', sa.String(255)),
        sa.Column('error_detail', sa.TEXT()),
        sa.Column('final', sa.Boolean),
        sa.Column('held', sa.Boolean),
        sa.Column('nodeset', sa.String(255)),
        sa.Column('ref_id', sa.Integer),
    )

    # Populate it with existing values, but get the ref_id for the
    # build from the buildset_ref table (we know that currently a
    # buildset is associated with exactly one ref, so we can use that
    # ref to associate its builds).
    statement = f"""
        insert into {prefixed_build_new}
            select
                {prefixed_build}.id,
                {prefixed_build}.buildset_id,
                {prefixed_build}.uuid,
                {prefixed_build}.job_name,
                {prefixed_build}.result,
                {prefixed_build}.start_time,
                {prefixed_build}.end_time,
                {prefixed_build}.voting,
                {prefixed_build}.log_url,
                {prefixed_build}.error_detail,
                {prefixed_build}.final,
                {prefixed_build}.held,
                {prefixed_build}.nodeset,
                {prefixed_buildset_ref}.ref_id
            from {prefixed_build} left join {prefixed_buildset_ref}
              on {prefixed_build}.buildset_id =
                 {prefixed_buildset_ref}.buildset_id
    """
    connection.execute(sa.text(statement))

    # Fix the sequence value since we wrote our own ids (postgres only)
    if dialect_name == 'postgresql':
        statement = f"""
            select setval(
                 '{prefixed_build_new}_id_seq',
                 COALESCE((SELECT MAX(id)+1 FROM {prefixed_build_new}), 1),
                 false)
        """
        connection.execute(sa.text(statement))

    # Add the foreign key indexes and constraits to our new table
    # first, to make sure we can before we drop the old one.
    with op.batch_alter_table(prefixed_build_new) as batch_op:
        batch_op.create_index(
            f'{prefixed_build_new}_buildset_id_idx',
            ['buildset_id'])
        batch_op.create_index(
            f'{prefixed_build_new}_ref_id_idx',
            ['ref_id'])

    # Mysql is quite slow about adding FK constraints.  But we're
    # pretty sure the data are valid (one of these constraints is
    # identical to one on the buildset table, and the other should be
    # identical to one on the buildset_ref table).
    # Disable FK checks from this point forward.
    if dialect_name in ('mysql', 'mariadb'):
        connection.execute(sa.text("set foreign_key_checks = 0"))

    statement = f"""
        alter table {prefixed_build_new}
        add constraint {prefixed_build_new}_buildset_id_fkey
            foreign key(buildset_id)
            references {prefixed_buildset_new} (id),
        add constraint {prefixed_build_new}_ref_id_fkey
            foreign key(ref_id)
            references {prefixed_ref} (id)
    """
    connection.execute(sa.text(statement))

    # After this point we expect everything else to succeed and
    # we can no longer roll back in mysql without data loss.


def rollback(connection, table_prefix):
    # This is our homemade best-effort rollback for mysql.
    prefixed_ref = table_prefix + REF_TABLE
    prefixed_build_new = table_prefix + BUILD_TABLE + '_new'
    prefixed_buildset_new = table_prefix + BUILDSET_TABLE + '_new'
    prefixed_buildset_ref = table_prefix + BUILDSET_REF_TABLE
    dialect_name = connection.engine.dialect.name

    connection.execute(sa.text(
        f"drop table if exists {prefixed_ref}"))
    connection.execute(sa.text(
        f"drop table if exists {prefixed_build_new}"))
    connection.execute(sa.text(
        f"drop table if exists {prefixed_buildset_new}"))
    connection.execute(sa.text(
        f"drop table if exists {prefixed_buildset_ref}"))
    if dialect_name in ('mysql', 'mariadb'):
        connection.execute(sa.text("set foreign_key_checks = 1"))


def upgrade2(connection, table_prefix):
    # This is the second part of the migration, past the point of no
    # return for mysql.
    prefixed_ref = table_prefix + REF_TABLE
    prefixed_build = table_prefix + BUILD_TABLE
    prefixed_build_new = table_prefix + BUILD_TABLE + '_new'
    prefixed_buildset = table_prefix + BUILDSET_TABLE
    prefixed_buildset_new = table_prefix + BUILDSET_TABLE + '_new'
    prefixed_artifact = table_prefix + ARTIFACT_TABLE
    prefixed_build_event = table_prefix + BUILD_EVENT_TABLE
    prefixed_provides = table_prefix + PROVIDES_TABLE
    quote = connection.engine.dialect.identifier_preparer.quote
    dialect_name = connection.engine.dialect.name

    # Temporarily drop the FK constraints that reference the old build
    # table. (This conditional is why we're renaming all the indexes
    # and constraints to be consistent across different backends).
    if dialect_name in ('mysql', 'mariadb'):
        op.drop_constraint(table_prefix + 'zuul_artifact_ibfk_1',
                           prefixed_artifact, 'foreignkey')
        op.drop_constraint(table_prefix + 'zuul_build_event_ibfk_1',
                           prefixed_build_event, 'foreignkey')
        op.drop_constraint(table_prefix + 'zuul_provides_ibfk_1',
                           prefixed_provides, 'foreignkey')
    elif dialect_name == 'postgresql':
        op.drop_constraint(table_prefix + 'zuul_artifact_build_id_fkey',
                           prefixed_artifact)
        op.drop_constraint(table_prefix + 'zuul_build_event_build_id_fkey',
                           prefixed_build_event)
        op.drop_constraint(table_prefix + 'zuul_provides_build_id_fkey',
                           prefixed_provides)
    else:
        raise Exception(f"Unsupported dialect {dialect_name}")

    # Drop the old table
    op.drop_table(prefixed_build)

    # Rename the table
    op.rename_table(prefixed_build_new, prefixed_build)

    # Rename the sequence and primary key (postgres only)
    if dialect_name == 'postgresql':
        statement = f"""
            alter sequence {prefixed_build_new}_id_seq
            rename to {prefixed_build}_id_seq;
        """
        connection.execute(sa.text(statement))
        rename_index(connection, prefixed_build_new,
                     f'{prefixed_build_new}_pkey',
                     f'{prefixed_build}_pkey')

    # Replace the indexes
    with op.batch_alter_table(prefixed_build) as batch_op:
        # This used to be named job_name_buildset_id_idx, let's
        # upgrade to our new naming scheme
        batch_op.create_index(
            f'{prefixed_build}_job_name_buildset_id_idx',
            ['job_name', 'buildset_id'])
        # Previously named uuid_buildset_id_idx
        batch_op.create_index(
            f'{prefixed_build}_uuid_buildset_id_idx',
            ['uuid', 'buildset_id'])

    # Rename indexes
    rename_index(connection, prefixed_build,
                 f'{prefixed_build_new}_buildset_id_idx',
                 f'{prefixed_build}_buildset_id_idx')
    rename_index(connection, prefixed_build,
                 f'{prefixed_build_new}_ref_id_idx',
                 f'{prefixed_build}_ref_id_idx')

    # Mysql does not support renaming constraints, so we drop and
    # re-add them.  We added them earlier to confirm that there were
    # no errors before dropping the original table (though in mysql we
    # did so with checks disabled, but postgres was able to validate
    # them.  We could have skipped the earlier add for mysql, but this
    # keeps the code simpler and more consistent, and is still fast).
    if dialect_name in ('mysql', 'mariadb'):
        statement = f"""
        alter table {prefixed_build}
        drop foreign key {prefixed_build_new}_buildset_id_fkey,
        drop foreign key {prefixed_build_new}_ref_id_fkey,
        add constraint {prefixed_build}_buildset_id_fkey
            foreign key(buildset_id)
            references {prefixed_buildset_new} (id),
        add constraint {prefixed_build}_ref_id_fkey
            foreign key(ref_id)
            references {prefixed_ref} (id)
        """
    elif dialect_name == 'postgresql':
        statement = f"""
        alter table {prefixed_build}
        drop constraint {prefixed_build_new}_buildset_id_fkey,
        drop constraint {prefixed_build_new}_ref_id_fkey,
        add constraint {prefixed_build}_buildset_id_fkey
            foreign key(buildset_id)
            references {prefixed_buildset_new} (id),
        add constraint {prefixed_build}_ref_id_fkey
            foreign key(ref_id)
            references {prefixed_ref} (id)
        """
    else:
        raise Exception(f"Unsupported dialect {dialect_name}")
    connection.execute(sa.text(statement))

    # Re-add the referencing FK constraints
    op.create_foreign_key(
        f'{prefixed_artifact}_build_id_fkey',
        prefixed_artifact,
        prefixed_build,
        ['build_id'], ['id'])
    op.create_foreign_key(
        f'{prefixed_build_event}_build_id_fkey',
        prefixed_build_event,
        prefixed_build,
        ['build_id'], ['id'])
    op.create_foreign_key(
        f'{prefixed_provides}_build_id_fkey',
        prefixed_provides,
        prefixed_build,
        ['build_id'], ['id'])

    # Rename some indexes for a consistent naming scheme
    rename_index(connection, prefixed_artifact,
                 f'{table_prefix}artifact_build_id_idx',
                 f'{prefixed_artifact}_build_id_idx')
    rename_index(connection, prefixed_build_event,
                 f'{table_prefix}build_event_build_id_idx',
                 f'{prefixed_build_event}_build_id_idx')
    rename_index(connection, prefixed_provides,
                 f'{table_prefix}provides_build_id_idx',
                 f'{prefixed_provides}_build_id_idx')

    # Rename the buildset table
    op.drop_table(prefixed_buildset)
    op.rename_table(prefixed_buildset_new, prefixed_buildset)

    # Rename the sequence and primary key (postgres only)
    if dialect_name == 'postgresql':
        statement = f"""
            alter sequence {prefixed_buildset_new}_id_seq
            rename to {prefixed_buildset}_id_seq;
        """
        connection.execute(sa.text(statement))
        rename_index(connection, prefixed_build_new,
                     f'{prefixed_buildset_new}_pkey',
                     f'{prefixed_buildset}_pkey')

    # Drop the columns that are no longer used in one statement for
    # efficiency (alembic doesn't have a way to do this).
    statement = f"""alter table {prefixed_buildset}
        drop column project,
        drop column {quote('change')},
        drop column patchset,
        drop column ref,
        drop column ref_url,
        drop column oldrev,
        drop column newrev,
        drop column branch
    """
    connection.execute(sa.text(statement))

    # Replace the only remaining buildset index
    op.create_index(
        f'{prefixed_buildset}_uuid_idx',
        prefixed_buildset,
        ['uuid'])

    # Re-enable FK checks for mysql
    if dialect_name in ('mysql', 'mariadb'):
        connection.execute(sa.text("set foreign_key_checks = 1"))


def upgrade(table_prefix=''):
    # The actual upgrade method, simplified for exception/rollback
    # handling.
    connection = op.get_bind()
    dialect_name = connection.engine.dialect.name
    if dialect_name not in ('mysql', 'mariadb', 'postgresql'):
        raise Exception(f"Unsupported dialect {dialect_name}")

    log = logging.getLogger('zuul.SQLMigration')
    try:
        upgrade1(connection, table_prefix)
    except Exception:
        try:
            if dialect_name in ('mysql', 'mariadb'):
                log.error("Early error in schema migration, rolling back")
                rollback(connection, table_prefix)
        except Exception:
            log.exception("Error in migration rollback:")
        raise
    upgrade2(connection, table_prefix)


def downgrade():
    raise Exception("Downgrades not supported")



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import multiprocessing
import os

from zuul.executor.sensors import SensorInterface
from zuul.lib.config import get_default


class CPUSensor(SensorInterface):
    log = logging.getLogger("zuul.executor.sensor.cpu")

    def __init__(self, statsd, base_key, config=None):
        super().__init__(statsd, base_key)
        load_multiplier = float(get_default(config, 'executor',
                                            'load_multiplier', '2.5'))
        self.max_load_avg = multiprocessing.cpu_count() * load_multiplier

    def isOk(self):
        load_avg = os.getloadavg()[0]

        if self.statsd:
            self.statsd.gauge(self.base_key + '.load_average',
                              int(load_avg * 100))

        if load_avg > self.max_load_avg:
            return False, "high system load {} > {}".format(
                load_avg, self.max_load_avg)

        return True, "load avg {} <= {}".format(load_avg, self.max_load_avg)



# Copyright 2012 Hewlett-Packard Development Company, L.P.
# Copyright 2016 Red Hat, Inc.
# Copyright 2021-2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import random
import hashlib

FIXTURE_DIR = os.path.join(os.path.dirname(__file__), 'fixtures')


def random_sha1():
    return hashlib.sha1(str(random.random()).encode('ascii')).hexdigest()



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import multiprocessing

from zuul.executor.sensors import SensorInterface
from zuul.lib.config import get_default


class StartingBuildsSensor(SensorInterface):
    log = logging.getLogger("zuul.executor.sensor.startingbuilds")

    def __init__(self, statsd, base_key, executor, max_load_avg, config=None):
        super().__init__(statsd, base_key)
        self.executor = executor

        coefficient = 2 if multiprocessing.cpu_count() <= 4 else 1
        max_default = int(max_load_avg * coefficient)

        self.max_starting_builds = get_default(
            config, 'executor', 'max_starting_builds', max_default) \
            if config is not None else max_default

        self.min_starting_builds = min(
            max(int(multiprocessing.cpu_count() / 2), 1),
            self.max_starting_builds)

    def _getStartingBuilds(self):
        starting_builds = 0
        for worker in self.executor.job_workers.values():
            if not worker.started:
                starting_builds += 1
        return starting_builds

    def _getRunningBuilds(self):
        return len(self.executor.job_workers)

    def _getPausedBuilds(self):
        paused_builds = 0
        for worker in self.executor.job_workers.values():
            if not worker.paused:
                paused_builds += 1
        return paused_builds

    def isOk(self):
        starting_builds = self._getStartingBuilds()
        running_builds = self._getRunningBuilds()
        paused_builds = self._getPausedBuilds()
        max_starting_builds = max(
            self.max_starting_builds - running_builds,
            self.min_starting_builds)
        if self.statsd:
            self.statsd.gauge(self.base_key + '.paused_builds',
                              paused_builds)
            self.statsd.gauge(self.base_key + '.running_builds',
                              running_builds)
            self.statsd.gauge(self.base_key + '.starting_builds',
                              starting_builds)

        if starting_builds >= max_starting_builds:
            return False, "too many starting builds {} >= {}".format(
                starting_builds, max_starting_builds)

        return True, "starting builds {} <= {}".format(
            starting_builds, max_starting_builds)



# Copyright 2021-2025 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from zuul import model
from zuul.provider import statemachine


class AzureProviderNode(model.ProviderNode, subclass_id="azure"):
    def __init__(self):
        super().__init__()
        self._set(
            azure_vm_id=None,
        )

    def getDriverData(self):
        return dict(
            azure_vm_id=self.azure_vm_id,
        )


class AzureInstance(statemachine.Instance):
    def __init__(self, region, vm, quota, nic=None, public_ipv4=None,
                 public_ipv6=None):
        super().__init__()
        self.azure_vm_id = vm and vm['id'] or None
        self.metadata = vm.get('tags', {})
        self.private_ipv4 = None
        self.private_ipv6 = None
        self.public_ipv4 = None
        self.public_ipv6 = None
        self.quota = quota

        if nic:
            for ip_config_data in nic['properties']['ipConfigurations']:
                ip_config_prop = ip_config_data['properties']
                if ip_config_prop['privateIPAddressVersion'] == 'IPv4':
                    self.private_ipv4 = ip_config_prop['privateIPAddress']
                if ip_config_prop['privateIPAddressVersion'] == 'IPv6':
                    self.private_ipv6 = ip_config_prop['privateIPAddress']

        if public_ipv4:
            self.public_ipv4 = public_ipv4['properties'].get('ipAddress')
        if public_ipv6:
            self.public_ipv6 = public_ipv6['properties'].get('ipAddress')

        self.interface_ip = (self.public_ipv4 or self.public_ipv6 or
                             self.private_ipv4 or self.private_ipv6)
        self.cloud = 'Azure'
        self.region = vm['location']
        if len(vm.get('zones', [])) > 0:
            self.az = vm['zones'][0]
        else:
            self.az = ''

    def getQuotaInformation(self):
        return self.quota

    @property
    def external_id(self):
        return f'vm={self.azure_vm_id}'


class AzureResource(statemachine.Resource):
    TYPE_INSTANCE = 'instance'
    TYPE_NIC = 'nic'
    TYPE_PIP = 'pip'
    TYPE_DISK = 'disk'
    TYPE_IMAGE = 'image'

    def __init__(self, metadata, type, id):
        super().__init__(metadata, type)
        self.id = id

    @property
    def unique_id(self):
        return '-'.join([self.type, self.id])



# Copyright 2019 Red Hat, Inc.
# Copyright 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import time
import logging

import voluptuous as v

from zuul.reporter import BaseReporter


class ElasticsearchReporter(BaseReporter):
    name = 'elasticsearch'
    log = logging.getLogger("zuul.ElasticsearchReporter")

    def __init__(self, driver, connection, config):
        super(ElasticsearchReporter, self).__init__(driver, connection, config)
        self.index = self.config.get('index', 'zuul')
        self.index_vars = self.config.get('index-vars')
        self.index_returned_vars = self.config.get('index-returned-vars')

    def report(self, item, phase1=True, phase2=True):
        """Create an entry into a database."""
        if not phase1:
            return
        docs = []
        index = '%s.%s-%s' % (self.index, item.manager.tenant.name,
                              time.strftime("%Y.%m.%d"))
        changes = [
            {
                "project": change.project.name,
                "change": getattr(change, 'number', None),
                "patchset": getattr(change, 'patchset', None),
                "ref": getattr(change, 'ref', ''),
                "oldrev": getattr(change, 'oldrev', ''),
                "newrev": getattr(change, 'newrev', ''),
                "branch": getattr(change, 'branch', ''),
                "ref_url": change.url,
            }
            for change in item.changes
        ]
        buildset_doc = {
            "uuid": item.current_build_set.uuid,
            "build_type": "buildset",
            "tenant": item.manager.tenant.name,
            "pipeline": item.manager.pipeline.name,
            "changes": changes,
            "project": item.changes[0].project.name,
            "change": getattr(item.changes[0], 'number', None),
            "patchset": getattr(item.changes[0], 'patchset', None),
            "ref": getattr(item.changes[0], 'ref', ''),
            "oldrev": getattr(item.changes[0], 'oldrev', ''),
            "newrev": getattr(item.changes[0], 'newrev', ''),
            "branch": getattr(item.changes[0], 'branch', ''),
            "zuul_ref": item.current_build_set.ref,
            "ref_url": item.changes[0].url,
            "result": item.current_build_set.result,
            "message": self._formatItemReport(item, with_jobs=False)
        }

        for job in item.getJobs():
            build = item.current_build_set.getBuild(job)
            if not build:
                continue
            # Ensure end_time is defined
            if not build.end_time:
                build.end_time = time.time()
            # Ensure start_time is defined
            if not build.start_time:
                build.start_time = build.end_time

            (result, url) = item.formatJobResult(job)

            # Manage to set time attributes in buildset
            start_time = int(build.start_time)
            end_time = int(build.end_time)
            if ('start_time' not in buildset_doc or
                    buildset_doc['start_time'] > start_time):
                buildset_doc['start_time'] = start_time
            if ('end_time' not in buildset_doc or
                    buildset_doc['end_time'] < end_time):
                buildset_doc['end_time'] = end_time
            buildset_doc['duration'] = (
                buildset_doc['end_time'] - buildset_doc['start_time'])

            change = item.getChangeForJob(build.job)
            change_doc = {
                "project": change.project.name,
                "change": getattr(change, 'number', None),
                "patchset": getattr(change, 'patchset', None),
                "ref": getattr(change, 'ref', ''),
                "oldrev": getattr(change, 'oldrev', ''),
                "newrev": getattr(change, 'newrev', ''),
                "branch": getattr(change, 'branch', ''),
                "ref_url": change.url,
            }

            build_doc = {
                "uuid": build.uuid,
                "change": change_doc,
                "build_type": "build",
                "buildset_uuid": buildset_doc['uuid'],
                "job_name": build.job.name,
                "result": result,
                "start_time": str(start_time),
                "end_time": str(end_time),
                "duration": end_time - start_time,
                "voting": build.job.voting,
                "log_url": url,
            }

            # Extends the build doc with some buildset info
            for attr in (
                    'tenant', 'pipeline', 'project', 'change', 'patchset',
                    'ref', 'oldrev', 'newrev', 'branch'):
                build_doc[attr] = buildset_doc[attr]

            if self.index_vars:
                build_doc['job_vars'] = job.variables

            if self.index_returned_vars:
                rdata = build.result_data.copy()
                rdata.pop('zuul', None)
                build_doc['job_returned_vars'] = rdata

            docs.append(build_doc)

        docs.append(buildset_doc)
        self.connection.add_docs(docs, index)


def getSchema():
    el_reporter = v.Schema(
        v.Any(
            None,
            {
                'index': str,
                'index-vars': bool,
                'index-returned-vars': bool
            }
        )
    )
    return el_reporter



"""Add oldrev/newrev columns

Revision ID: ba4cdce9b18c
Revises: 5efb477fa963
Create Date: 2017-09-27 19:33:21.800198

"""

# revision identifiers, used by Alembic.
revision = 'ba4cdce9b18c'
down_revision = '5efb477fa963'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade(table_prefix=''):
    op.add_column(
        table_prefix + 'zuul_buildset', sa.Column('oldrev', sa.String(255)))
    op.add_column(
        table_prefix + 'zuul_buildset', sa.Column('newrev', sa.String(255)))


def downgrade():
    raise Exception("Downgrades not supported")



#!/usr/bin/env python
# Copyright 2012 Hewlett-Packard Development Company, L.P.
# Copyright 2013-2014 OpenStack Foundation
# Copyright 2021-2022 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import signal

import zuul.cmd
import zuul.merger.server


class Merger(zuul.cmd.ZuulDaemonApp):
    app_name = 'merger'
    app_description = 'A standalone Zuul merger.'

    def createParser(self):
        parser = super(Merger, self).createParser()
        self.addSubCommands(parser, zuul.merger.server.COMMANDS)
        return parser

    def exit_handler(self, signum, frame):
        self.merger.stop()
        self.merger.join()

    def run(self):
        self.handleCommands()

        self.configure_connections(sources=True)

        self.setup_logging('merger', 'log_config')

        self.merger = zuul.merger.server.MergeServer(
            self.config, self.connections)
        self.merger.start()

        if self.args.nodaemon:
            signal.signal(signal.SIGTERM, self.exit_handler)

        while True:
            try:
                self.merger.join()
                break
            except KeyboardInterrupt:
                print("Ctrl + C: asking merger to exit nicely...\n")
                self.exit_handler(signal.SIGINT, None)


def main():
    Merger().main()



# Copyright 2019 BMW Group
# Copyright 2021 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging


def get_annotated_logger(logger, event=None, *, build=None, request=None,
                         upload=None, iba=None):
    # Note(tobiash): When running with python 3.5 log adapters cannot be
    # stacked. We need to detect this case and modify the original one.
    if isinstance(logger, EventIdLogAdapter):
        extra = logger.extra
    else:
        extra = {}

    if event is not None:
        if hasattr(event, 'zuul_event_id'):
            extra['event_id'] = event.zuul_event_id
        else:
            extra['event_id'] = event

    if build is not None:
        extra['build'] = build

    if request is not None:
        extra['request'] = request

    if upload is not None:
        extra['image_upload'] = upload

    if iba is not None:
        extra['image_build_artifact'] = iba

    if isinstance(logger, EventIdLogAdapter):
        return logger

    return EventIdLogAdapter(logger, extra)


class EventIdLogAdapter(logging.LoggerAdapter):
    def process(self, msg, kwargs):
        msg, kwargs = super().process(msg, kwargs)
        extra = kwargs.get('extra', {})
        new_msg = []

        event_id = extra.get('event_id')
        if event_id is not None:
            new_msg.append('[e: %s]' % event_id)
        build = extra.get('build')
        if build is not None:
            new_msg.append('[build: %s]' % build)
        request = extra.get('request')
        if request is not None:
            new_msg.append('[req: %s]' % request)
        upload = extra.get('image_upload')
        if upload is not None:
            new_msg.append('[upload: %s]' % upload)
        iba = extra.get('image_build_artifact')
        if iba is not None:
            new_msg.append('[iba: %s]' % iba)

        new_msg.append(msg)
        msg = ' '.join(new_msg)
        return msg, kwargs

    def addHandler(self, *args, **kw):
        return self.logger.addHandler(*args, **kw)


class MultiLineFormatter(logging.Formatter):
    def format(self, record):
        rec = super().format(record)
        ret = []
        # Save the existing message and re-use this record object to
        # format each line.
        saved_msg = record.message
        for i, line in enumerate(rec.split('\n')):
            if i:
                record.message = '  ' + line
                ret.append(self.formatMessage(record))
            else:
                ret.append(line)
        # Restore the message
        record.message = saved_msg
        return '\n'.join(ret)



# Copyright 2020 BMW Group
# Copyright 2021-2025 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import enum
import functools
import json
import logging
import threading
import time
import uuid
import zlib
from collections import namedtuple
from collections.abc import Iterable
from contextlib import suppress

from kazoo.exceptions import NoNodeError
from kazoo.protocol.states import EventType

from zuul import model
from zuul.lib.collections import DefaultKeyDict
from zuul.lib.logutil import get_annotated_logger
from zuul.zk import ZooKeeperSimpleBase, sharding
from zuul.zk.election import SessionAwareElection, RendezvousElection

RESULT_EVENT_TYPE_MAP = {
    "BuildCompletedEvent": model.BuildCompletedEvent,
    "BuildPausedEvent": model.BuildPausedEvent,
    "BuildStartedEvent": model.BuildStartedEvent,
    "BuildStatusEvent": model.BuildStatusEvent,
    "FilesChangesCompletedEvent": model.FilesChangesCompletedEvent,
    "MergeCompletedEvent": model.MergeCompletedEvent,
    "NodesProvisionedEvent": model.NodesProvisionedEvent,
    # MODEL_API <= 32
    # Kept for backward compatibility; semaphore release events
    # are now processed in the management event queue.
    "SemaphoreReleaseEvent": model.SemaphoreReleaseEvent,
}

MANAGEMENT_EVENT_TYPE_MAP = {
    "DequeueEvent": model.DequeueEvent,
    "EnqueueEvent": model.EnqueueEvent,
    "PromoteEvent": model.PromoteEvent,
    "ReconfigureEvent": model.ReconfigureEvent,
    "SemaphoreReleaseEvent": model.SemaphoreReleaseEvent,
    "TenantReconfigureEvent": model.TenantReconfigureEvent,
    "PipelinePostConfigEvent": model.PipelinePostConfigEvent,
    "PipelineSemaphoreReleaseEvent": model.PipelineSemaphoreReleaseEvent,
}

# /zuul/events/tenant   TENANT_ROOT
#   /{tenant}           TENANT_NAME_ROOT
#     /state            TENANT_EVENT_STATE
#     /management       TENANT_MANAGEMENT_ROOT
#       /queue          TENANT_MANAGEMENT_QUEUE
#       /data           [side channel data]
#     /trigger          TENANT_TRIGGER_ROOT
#       /queue          TENANT_TRIGGER_QUEUE
#       /data           [side channel data]
#     /pipelines        PIPELINE_ROOT
#       /{pipeline}     PIPELINE_NAME_ROOT
#         /management   PIPELINE_MANAGEMENT_ROOT
#           /queue      PIPELINE_MANAGEMENT_QUEUE
#           /data       [side channel data]
#         /trigger      PIPELINE_TRIGGER_ROOT
#           /queue      PIPELINE_TRIGGER_QUEUE
#           /data       [side channel data]
#         /result       PIPELINE_RESULT_ROOT
#           /queue      PIPELINE_RESULT_QUEUE
#           /data       [side channel data]

TENANT_ROOT = "/zuul/events/tenant"
TENANT_NAME_ROOT = TENANT_ROOT + "/{tenant}"
TENANT_MANAGEMENT_ROOT = TENANT_NAME_ROOT + "/management"
TENANT_MANAGEMENT_QUEUE = TENANT_MANAGEMENT_ROOT + "/queue"
TENANT_TRIGGER_ROOT = TENANT_NAME_ROOT + "/trigger"
TENANT_TRIGGER_QUEUE = TENANT_TRIGGER_ROOT + "/queue"
PIPELINE_ROOT = TENANT_NAME_ROOT + "/pipelines"
PIPELINE_NAME_ROOT = PIPELINE_ROOT + "/{pipeline}"
PIPELINE_MANAGEMENT_ROOT = PIPELINE_NAME_ROOT + "/management"
PIPELINE_MANAGEMENT_QUEUE = PIPELINE_MANAGEMENT_ROOT + "/queue"
PIPELINE_TRIGGER_ROOT = PIPELINE_NAME_ROOT + "/trigger"
PIPELINE_TRIGGER_QUEUE = PIPELINE_TRIGGER_ROOT + "/queue"
PIPELINE_RESULT_ROOT = PIPELINE_NAME_ROOT + "/result"
PIPELINE_RESULT_QUEUE = PIPELINE_RESULT_ROOT + "/queue"
TENANT_EVENT_STATE = TENANT_NAME_ROOT + "/state"

CONNECTION_ROOT = "/zuul/events/connection"

# This is the path to the serialized from of the event in ZK (along
# with the version when it was read (which should not change since
# events are immutable in queue)).  When processing of the event is
# complete, this is the path that should be deleted in order to
# acknowledge it and prevent re-processing.  Instances of this are
# dynamically created and attached to de-serialized Event instances.
EventAckRef = namedtuple("EventAckRef", ("path", "version"))

UNKNOWN_ZVERSION = -1


class EventPrefix(enum.Enum):
    MANAGEMENT = "100"
    RESULT = "200"
    TRIGGER = "300"


class EventWatcher(ZooKeeperSimpleBase):

    log = logging.getLogger("zuul.EventWatcher")

    def __init__(self, client, callback):
        super().__init__(client)
        self.callback = callback
        self.watched_tenants = set()
        self.watched_pipelines = set()
        self.tenant_state = DefaultKeyDict(
            lambda tenant_name: model.TenantEventState(
                TENANT_EVENT_STATE.format(tenant=tenant_name)))
        self.kazoo_client.ensure_path(TENANT_ROOT)
        self.kazoo_client.ChildrenWatch(TENANT_ROOT, self._tenantWatch)

    def _makePipelineWatcher(self, tenant_name):
        def watch(children, event=None):
            return self._pipelineWatch(tenant_name, children)
        return watch

    def _tenantWatch(self, tenants):
        for tenant_name in tenants:
            if tenant_name in self.watched_tenants:
                continue

            if self.callback:
                # only set these watches if we're in the scheduler context
                for path in (TENANT_MANAGEMENT_QUEUE,
                             TENANT_TRIGGER_QUEUE):
                    path = path.format(tenant=tenant_name)
                    self.kazoo_client.ensure_path(path)
                    self.kazoo_client.ChildrenWatch(
                        path, self._eventWatch, send_event=True)

                pipelines_path = PIPELINE_ROOT.format(tenant=tenant_name)
                self.kazoo_client.ensure_path(pipelines_path)
                self.kazoo_client.ChildrenWatch(
                    pipelines_path, self._makePipelineWatcher(tenant_name))

            # both scheduler and zuul-web are interested in these
            state_path = TENANT_EVENT_STATE.format(tenant=tenant_name)
            self.kazoo_client.DataWatch(
                state_path,
                self._makeStateWatcher(tenant_name))
            self.watched_tenants.add(tenant_name)

    def _pipelineWatch(self, tenant_name, pipelines):
        # Remove pipelines that no longer exists from the watch list so
        # we re-register the children watch in case the pipeline is
        # added again.
        for watched_tenant, pipeline_name in list(self.watched_pipelines):
            if watched_tenant != tenant_name:
                continue
            if pipeline_name in pipelines:
                continue
            with suppress(KeyError):
                self.watched_pipelines.remove((tenant_name, pipeline_name))

        for pipeline_name in pipelines:
            key = (tenant_name, pipeline_name)
            if key in self.watched_pipelines:
                continue

            for path in (PIPELINE_MANAGEMENT_QUEUE,
                         PIPELINE_TRIGGER_QUEUE,
                         PIPELINE_RESULT_QUEUE):
                path = path.format(tenant=tenant_name,
                                   pipeline=pipeline_name)

                self.kazoo_client.ensure_path(path)
                self.kazoo_client.ChildrenWatch(
                    path, self._eventWatch, send_event=True)
            self.watched_pipelines.add(key)

    def _eventWatch(self, event_list, event=None):
        if event is None:
            # Handle initial call when the watch is created. If there are
            # already events in the queue we trigger the callback.
            if event_list:
                self.callback()
        elif event.type == EventType.CHILD:
            self.callback()

    def _makeStateWatcher(self, tenant_name):
        def watch(data=None, event=None):
            return self._stateWatch(tenant_name, data, event)
        return watch

    def _stateWatch(self, tenant_name,
                    data=None, stat=None):
        if data is None:
            self.tenant_state[tenant_name].reset()
        else:
            self.tenant_state[tenant_name]._updateFromRaw(
                None, data, stat, None)
            if self.callback:
                self.callback()


class ZooKeeperEventQueue(ZooKeeperSimpleBase, Iterable):
    """Abstract API for events via ZooKeeper

    The lifecycle of a global (not pipeline-specific) event is:

    * Serialized form of event added to ZK queue.

    * During queue processing, events are de-serialized and
      AbstractEvent subclasses are instantiated.  An EventAckRef is
      associated with the event instance in order to maintain the link
      to the serialized form.

    * When event processing is complete, the EventAckRef is used to
      delete the original event.  If the event requires a result
      (e.g., a management event that returns data) the result will be
      written to a pre-determined location.  A future can watch for
      the result to appear at that location.

    Pipeline specific events usually start out as global events, but
    upon processing, may be forwarded to pipeline-specific queues.  In
    these cases, the original event will be deleted as above, and a
    new, identical event will be created in the pipeline-specific
    queue.  If the event expects a result, no result will be written
    upon forwarding; the result will only be written when the
    forwarded event is processed.

    """

    log = logging.getLogger("zuul.ZooKeeperEventQueue")

    # How often to update our internal cache of the queue size
    CACHED_COUNT_INTERVAL = 10
    # The maximum number of events in the queue; more than this will
    # be discarded.
    MAX_QUEUE = 90000

    def __init__(self, client, queue_root):
        super().__init__(client)
        self.queue_root = queue_root
        self.event_root = f'{queue_root}/queue'
        self.data_root = f'{queue_root}/data'
        self._cached_count = 0
        self._cached_count_time = 0
        self.initialize()

    def initialize(self):
        self.kazoo_client.ensure_path(self.event_root)
        self.kazoo_client.ensure_path(self.data_root)

    def _listEvents(self):
        return self.kazoo_client.get_children(self.event_root)

    def __len__(self):
        try:
            data, stat = self.kazoo_client.get(self.event_root)
            return stat.children_count
        except NoNodeError:
            return 0

    def hasEvents(self):
        return bool(len(self))

    def ack(self, event):
        # Event.ack_ref is an EventAckRef, previously attached to an
        # event object when it was de-serialized.
        if not event.ack_ref:
            raise RuntimeError("Cannot ack event %s without reference", event)
        if not self._remove(event.ack_ref.path, event.ack_ref.version):
            self.log.warning("Event %s was already acknowledged", event)

    def eventIdFromAckRef(self, ack_ref):
        _, _, event_id = ack_ref.path.rpartition("/")
        return event_id

    def _put(self, data, updater=None):
        # Event data can be large, so we want to shard it.  But events
        # also need to be atomic (we don't want an event listener to
        # start processing a half-stored event).  A natural solution
        # is to write the sharded data along with the event.  However,
        # our events are sequence nodes in order to maintain ordering,
        # and we can't write shard nodes under the event without
        # creating the event first.  We can't use transactions because
        # they are subject to the same size limit.  To resolve this,
        # we store the event data in two places: the event itself and
        # associated metadata are in the event queue as a single
        # sequence node.  The event data are stored in a separate tree
        # under a UUID.  We have to use a UUID because we don't know
        # the sequence node when we write it.  The event metadata
        # includes the UUID of the data.  We call the separated data
        # "side channel data" to indicate it's stored outside the main
        # event queue.
        #
        # To make the API simpler to work with, we assume "event_data"
        # contains the bulk of the data.  We extract it here into the
        # side channel data, then in _iterEvents we re-constitute it
        # into the dictionary.
        #
        # If `updater` is supplied, it is a RequestUpdater instance, and
        # we are engaged in a cooperative transaction with a job
        # request.

        # Add a trailing /q for our sequence nodes; transaction
        # sequence nodes need a character after / otherwise they drop
        # the /.
        event_path = f"{self.event_root}/q"

        if updater and not updater.preRun():
            # Don't even try the update
            return

        # Protect against a ZK error if the queue is too large; ZK
        # will let us add children, but we won't be able to get the
        # list of children if there are more than about 95k of them.
        # We implement an internal limit of 90k events.  If the queue
        # has more than that number, we discard events.  In order to
        # avoid adding an extra ZK "stat" call before every "create",
        # we cache the queue size, and update it only every 10
        # seconds.  However, if the cached size exceeds 50% of the
        # max, then we update it on every request, to reduce the
        # chances of going over the limit.  Additionally, we increment
        # our cached value every time we put an item in the queue.  If
        # there is a large rush of 50k events that would push us over
        # the limit in less than 10 seconds, they are likely to come
        # from the same place, which means our internal counter would
        # detect that and start issuing the stat call more often.
        # Nothing decrements the counter, but if it gets too high, we
        # will issue a stat call and override the cache.
        now = time.monotonic()
        if ((now - self._cached_count_time > self.CACHED_COUNT_INTERVAL) or
            self._cached_count > (self.MAX_QUEUE / 2)):
            self._cached_count = len(self)
            self._cached_count_time = now
        if self._cached_count >= self.MAX_QUEUE:
            self.log.warning("Discarding event %s due to queue length %s",
                             data, self._cached_count)
            return
        self._cached_count += 1

        # Write the side channel data here under the assumption that
        # the transaction will proceed.  If the transaction fails, it
        # will be cleaned up after about 5 minutes.
        side_channel_data = None
        size_limit = sharding.NODE_BYTE_SIZE_LIMIT
        if updater:
            # If we are in a transaction, leave enough room to share.
            size_limit /= 2
        encoded_data = json.dumps(data, sort_keys=True).encode("utf-8")
        if (len(encoded_data) > size_limit
            and 'event_data' in data):
            # Get a unique data node
            data_id = str(uuid.uuid4())
            data_root = f'{self.data_root}/{data_id}'
            side_channel_data = json.dumps(data['event_data'],
                                           sort_keys=True).encode("utf-8")
            data = data.copy()
            del data['event_data']
            data['event_data_path'] = data_root
            encoded_data = json.dumps(data, sort_keys=True).encode("utf-8")

            with sharding.BufferedShardWriter(
                    self.kazoo_client, data_root) as stream:
                stream.write(zlib.compress(side_channel_data))

        if updater is None:
            return self.kazoo_client.create(
                event_path, encoded_data, sequence=True)

        # The rest of the method is the updater case.  Start a
        # transaction.
        transaction = self.kazoo_client.transaction()

        # Add transaction tasks.
        transaction.create(event_path, encoded_data, sequence=True)
        updater.run(transaction)

        # Commit the transaction and process the results.  `results`
        # is an array of either exceptions or return values
        # corresponding to the operations in order.
        result = transaction.commit()
        if isinstance(result[0], Exception):
            raise result[0]

        updater.postRun(result[1])
        return result[0]

    def _iterEvents(self, event_id_offset=None):
        try:
            # We need to sort this ourself, since Kazoo doesn't guarantee any
            # ordering of the returned children.
            events = sorted(self._listEvents())
        except NoNodeError:
            return

        for event_id in events:
            if event_id_offset and event_id <= event_id_offset:
                continue
            path = "/".join((self.event_root, event_id))

            # Load the event metadata
            data, zstat = self.kazoo_client.get(path)
            try:
                event = json.loads(data)
            except json.JSONDecodeError as e:
                self.log.error("Malformed event data in %s: %s", path, e)
                self._remove(path)
                continue

            # Load the event data (if present); if that fails, the
            # event is corrupt; delete and continue.
            side_channel_path = event.get('event_data_path')
            if side_channel_path:
                try:
                    with sharding.BufferedShardReader(
                            self.kazoo_client, side_channel_path) as stream:
                        side_channel_data = zlib.decompress(stream.read())
                except NoNodeError:
                    self.log.exception("Side channel data for %s "
                                       "not found at %s",
                                       path, side_channel_path)
                    self._remove(path)
                    continue

                try:
                    event_data = json.loads(side_channel_data)
                except json.JSONDecodeError:
                    self.log.exception("Malformed side channel "
                                       "event data in %s",
                                       side_channel_path)
                    self._remove(path)
                    continue
                event['event_data'] = event_data

            yield event, EventAckRef(path, zstat.version), zstat

    def _remove(self, path, version=UNKNOWN_ZVERSION):
        try:
            # Find the side channel path

            side_channel_path = None
            data, zstat = self.kazoo_client.get(path)
            try:
                event = json.loads(data)
                side_channel_path = event.get('event_data_path')
            except json.JSONDecodeError:
                pass

            if side_channel_path:
                with suppress(NoNodeError):
                    self.kazoo_client.delete(side_channel_path, recursive=True)

            self.kazoo_client.delete(path, version=version, recursive=True)
            return True
        except NoNodeError:
            return False

    def _findLostSideChannelData(self, age):
        # Get data nodes which are older than the specified age (we
        # don't want to delete nodes which are just being written
        # slowly).
        # Convert to MS
        now = int(time.time() * 1000)
        age = age * 1000
        data_nodes = set()
        for data_id in self.kazoo_client.get_children(self.data_root):
            data_path = f'{self.data_root}/{data_id}'
            data_zstat = self.kazoo_client.exists(data_path)
            if now - data_zstat.mtime > age:
                data_nodes.add(data_path)

        # If there are no candidate data nodes, we don't need to
        # filter them by known events.
        if not data_nodes:
            return data_nodes

        events = set(self._listEvents())

        for event_id in events:
            # Load the event metadata
            path = "/".join((self.event_root, event_id))
            data, zstat = self.kazoo_client.get(path)
            try:
                event = json.loads(data)
            except json.JSONDecodeError:
                self.log.exception("Malformed event data in %s", path)
                self._remove(path)
                continue

            side_channel_path = event.get('event_data_path')
            if side_channel_path:
                data_nodes.discard(side_channel_path)
        return data_nodes

    def cleanup(self, age=300):
        try:
            for path in self._findLostSideChannelData(age):
                try:
                    self.log.error("Removing side channel data: %s", path)
                    self.kazoo_client.delete(path, recursive=True)
                except Exception:
                    self.log.exception(
                        "Unable to delete side channel data %s", path)
        except Exception:
            self.log.exception("Error cleaning up event queue %s", self)


class EventResultFuture(ZooKeeperSimpleBase):
    """Base class for result futures for different events."""

    log = logging.getLogger("zuul.EventResultFuture")

    def __init__(self, client, result_path):
        super().__init__(client)
        self._result_path = result_path
        self._wait_event = threading.Event()
        self.kazoo_client.DataWatch(self._result_path, self._resultCallback)
        self.data = {}

    def _resultCallback(self, data=None, stat=None):
        if data is None:
            # Igore events w/o any data
            return None
        self._wait_event.set()
        # Stop the watch if we got a result
        return False

    def _read(self):
        # Internal method to read the result data; may be overridden
        # by subclasses.
        return self.kazoo_client.get(self._result_path)[0]

    def _delete(self):
        with suppress(NoNodeError):
            self.kazoo_client.delete(self._result_path, recursive=True)

    def wait(self, timeout=None):
        """Wait until the result for this event has been written."""

        # Note that due to event forwarding, the only way to confirm
        # that an event has been processed is to check for a result;
        # the original event may have been deleted when forwaded to a
        # different queue.
        if not self._wait_event.wait(timeout):
            return False
        try:
            try:
                data = self._read()
                self.data = json.loads(data.decode("utf-8"))
            except json.JSONDecodeError:
                self.log.exception(
                    "Malformed result data in %s", self._result_path
                )
                raise
        finally:
            self._delete()
        return True


class JobResultFuture(EventResultFuture):

    log = logging.getLogger("zuul.JobResultFuture")

    def __init__(self, client, request_path, result_path, waiter_path):
        super().__init__(client, result_path)

        self.request_path = request_path
        self._waiter_path = waiter_path
        self._result_data_path = None
        self.merged = None
        self.updated = None
        self.commit = None
        self.files = None
        self.repo_state = None
        self.item_in_branches = None

    def __repr__(self):
        return f"<JobResultFuture request_path={self.request_path}>"

    def _read(self):
        result_node = self.kazoo_client.get(self._result_path)[0]
        result = json.loads(result_node)
        self._result_data_path = result['result_data_path']
        with sharding.BufferedShardReader(
                self.kazoo_client, self._result_data_path) as stream:
            return zlib.decompress(stream.read())

    def _delete(self):
        super()._delete()
        if self._result_data_path:
            with suppress(NoNodeError):
                self.kazoo_client.delete(self._result_data_path,
                                         recursive=True)

    def wait(self, timeout=None):
        try:
            res = super().wait(timeout)
        finally:
            with suppress(NoNodeError):
                self.kazoo_client.delete(self._waiter_path)
        self.merged = self.data.get("merged", False)
        self.updated = self.data.get("updated", False)
        self.commit = self.data.get("commit")
        self.files = self.data.get("files", {})
        self.repo_state = self.data.get("repo_state", {})
        self.item_in_branches = self.data.get("item_in_branches", [])
        return res

    def cancel(self):
        # Remove our waiter node so that if a result is ever reported,
        # it will be garbage collected.
        with suppress(NoNodeError):
            self.kazoo_client.delete(self._waiter_path)


class ManagementEventResultFuture(EventResultFuture):
    """Returned when a management event is put into a queue."""

    log = logging.getLogger("zuul.ManagementEventResultFuture")

    def __init__(self, client, result_path: str):
        super().__init__(client, result_path)

    def wait(self, timeout=None):
        res = super().wait(timeout)
        tb = self.data.get("traceback")
        if tb is not None:
            # TODO: raise some kind of ManagementEventException here
            raise RuntimeError(tb)
        return res


class ManagementEventQueue(ZooKeeperEventQueue):
    """Management events via ZooKeeper"""

    RESULTS_ROOT = "/zuul/results/management"

    log = logging.getLogger("zuul.ManagementEventQueue")

    def put(self, event, needs_result=True):
        result_path = None
        # If this event is forwarded it might have a result ref that
        # we need to forward.
        if event.result_ref:
            result_path = event.result_ref
        elif needs_result:
            result_path = "/".join((self.RESULTS_ROOT, str(uuid.uuid4())))

        data = {
            "event_type": type(event).__name__,
            "event_data": event.toDict(),
            "result_path": result_path,
        }
        if needs_result and not event.result_ref:
            # The event was not forwarded, create the result ref
            self.kazoo_client.create(result_path, None,
                                     makepath=True, ephemeral=True)
        self._put(data)
        if needs_result and result_path:
            return ManagementEventResultFuture(self.client, result_path)
        return None

    def __iter__(self):
        event_list = []
        for data, ack_ref, zstat in self._iterEvents():
            try:
                event_class = MANAGEMENT_EVENT_TYPE_MAP[data["event_type"]]
                event_data = data["event_data"]
                result_path = data["result_path"]
            except KeyError:
                self.log.warning("Malformed event found: %s", data)
                self._remove(ack_ref.path, ack_ref.version)
                continue
            event = event_class.fromDict(event_data)
            event.ack_ref = ack_ref
            event.result_ref = result_path
            # Initialize the logical timestamp if not valid
            if event.zuul_event_ltime is None:
                event.zuul_event_ltime = zstat.creation_transaction_id

            try:
                other_event = event_list[event_list.index(event)]
            except ValueError:
                other_event = None
            if other_event:
                if isinstance(other_event, model.TenantReconfigureEvent):
                    other_event.merge(event)
                    continue
            event_list.append(event)
        yield from event_list

    def ack(self, event):
        """Acknowledge the event (by deleting it from the queue)"""
        # Note: the result is reported first to ensure that the
        # originator of the event which may be waiting on a future
        # receives a result, or otherwise this event is considered
        # unprocessed and remains on the queue.
        self._reportResult(event)
        super().ack(event)
        if isinstance(event, model.TenantReconfigureEvent):
            for merged_event in event.merged_events:
                merged_event.traceback = event.traceback
                self._reportResult(merged_event)
                super().ack(merged_event)

    def _reportResult(self, event):
        if not event.result_ref:
            return

        result_data = {"traceback": event.traceback,
                       "timestamp": time.time()}
        try:
            self.kazoo_client.set(
                event.result_ref,
                json.dumps(result_data, sort_keys=True).encode("utf-8"),
            )
        except NoNodeError:
            self.log.warning(f"No result node found for {event}; "
                             "client may have disconnected")


class PipelineManagementEventQueue(ManagementEventQueue):
    log = logging.getLogger(
        "zuul.zk.event_queues.PipelineManagementEventQueue"
    )

    def __init__(self, client, tenant_name, pipeline_name):
        queue_root = PIPELINE_MANAGEMENT_ROOT.format(
            tenant=tenant_name,
            pipeline=pipeline_name)
        super().__init__(client, queue_root)

    @classmethod
    def createRegistry(cls, client):
        """Create a tenant/pipeline queue registry

        Returns a nested dictionary of:
          tenant_name -> pipeline_name -> EventQueue

        Queues are dynamically created with the originally supplied ZK
        client as they are accessed via the registry (so new tenants
        or pipelines show up automatically).

        """
        return DefaultKeyDict(lambda t: cls._createRegistry(client, t))

    @classmethod
    def _createRegistry(cls, client, tenant_name):
        return DefaultKeyDict(lambda p: cls(client, tenant_name, p))


class TenantManagementEventQueue(ManagementEventQueue):
    log = logging.getLogger("zuul.TenantManagementEventQueue")

    def __init__(self, client, tenant_name):
        queue_root = TENANT_MANAGEMENT_ROOT.format(
            tenant=tenant_name)
        super().__init__(client, queue_root)

    def ackWithoutResult(self, event):
        """
        Used to ack a management event when forwarding to a pipeline queue
        """
        super(ManagementEventQueue, self).ack(event)
        if isinstance(event, model.TenantReconfigureEvent):
            for merged_event in event.merged_events:
                super(ManagementEventQueue, self).ack(merged_event)

    @classmethod
    def createRegistry(cls, client):
        """Create a tenant queue registry

        Returns a dictionary of: tenant_name -> EventQueue

        Queues are dynamically created with the originally supplied ZK
        client as they are accessed via the registry (so new tenants
        show up automatically).

        """
        return DefaultKeyDict(lambda t: cls(client, t))


class PipelineResultEventQueue(ZooKeeperEventQueue):
    """Result events via ZooKeeper"""

    log = logging.getLogger("zuul.PipelineResultEventQueue")

    def __init__(self, client, tenant_name, pipeline_name):
        queue_root = PIPELINE_RESULT_ROOT.format(
            tenant=tenant_name,
            pipeline=pipeline_name)
        super().__init__(client, queue_root)

    @classmethod
    def createRegistry(cls, client):
        """Create a tenant/pipeline queue registry

        Returns a nested dictionary of:
          tenant_name -> pipeline_name -> EventQueue

        Queues are dynamically created with the originally supplied ZK
        client as they are accessed via the registry (so new tenants
        or pipelines show up automatically).

        """
        return DefaultKeyDict(lambda t: cls._createRegistry(client, t))

    @classmethod
    def _createRegistry(cls, client, tenant_name):
        return DefaultKeyDict(lambda p: cls(client, tenant_name, p))

    def put(self, event, updater=None):
        data = {
            "event_type": type(event).__name__,
            "event_data": event.toDict(),
        }
        self._put(data, updater)

    def __iter__(self):
        for data, ack_ref, _ in self._iterEvents():
            try:
                event_class = RESULT_EVENT_TYPE_MAP[data["event_type"]]
                event_data = data["event_data"]
            except KeyError:
                self.log.warning("Malformed event found: %s", data)
                self._remove(ack_ref.path, ack_ref.version)
                continue
            event = event_class.fromDict(event_data)
            event.ack_ref = ack_ref
            yield event


class TriggerEventQueue(ZooKeeperEventQueue):
    """Trigger events via ZooKeeper"""

    log = logging.getLogger("zuul.TriggerEventQueue")

    def __init__(self, client, queue_root, connections):
        self.connections = connections
        super().__init__(client, queue_root)

    def put(self, driver_name, event):
        data = {
            "driver_name": driver_name,
            "event_data": event.toDict(),
        }
        self._put(data)

    def put_supercede(self, event):
        data = {
            "event_type": "SupercedeEvent",
            "event_data": event.toDict(),
            "driver_name": None,
        }
        self._put(data)

    def __iter__(self):
        for data, ack_ref, zstat in self._iterEvents():
            try:
                if (data["driver_name"] is None and
                        data["event_type"] == "SupercedeEvent"):
                    event_class = model.SupercedeEvent
                else:
                    event_class = self.connections.getTriggerEventClass(
                        data["driver_name"]
                    )
                event_data = data["event_data"]
            except KeyError:
                self.log.warning("Malformed event found: %s", data)
                self._remove(ack_ref.path, ack_ref.version)
                continue
            event = event_class.fromDict(event_data)
            event.ack_ref = ack_ref
            event.driver_name = data["driver_name"]
            # Initialize the logical timestamp if not valid
            if event.zuul_event_ltime is None:
                event.zuul_event_ltime = zstat.creation_transaction_id
            yield event


class TenantTriggerEventQueue(TriggerEventQueue):
    log = logging.getLogger("zuul.TenantTriggerEventQueue")

    def __init__(self, client, connections, tenant_name):
        queue_root = TENANT_TRIGGER_ROOT.format(
            tenant=tenant_name)
        super().__init__(client, queue_root, connections)
        self.metadata = {}

    def _setQueueMetadata(self):
        encoded_data = json.dumps(
            self.metadata, sort_keys=True).encode("utf-8")
        self.kazoo_client.set(self.queue_root, encoded_data)

    def refreshMetadata(self):
        data, zstat = self.kazoo_client.get(self.queue_root)
        try:
            self.metadata = json.loads(data)
        except json.JSONDecodeError:
            self.metadata = {}

    @property
    def last_reconfigure_event_ltime(self):
        return self.metadata.get('last_reconfigure_event_ltime', -1)

    @last_reconfigure_event_ltime.setter
    def last_reconfigure_event_ltime(self, val):
        self.metadata['last_reconfigure_event_ltime'] = val
        self._setQueueMetadata()

    @classmethod
    def createRegistry(cls, client, connections):
        """Create a tenant queue registry

        Returns a dictionary of: tenant_name -> EventQueue

        Queues are dynamically created with the originally supplied ZK
        client as they are accessed via the registry (so new tenants
        show up automatically).

        """
        return DefaultKeyDict(lambda t: cls(client, connections, t))


class PipelineTriggerEventQueue(TriggerEventQueue):
    log = logging.getLogger("zuul.PipelineTriggerEventQueue")

    def __init__(self, client, tenant_name, pipeline_name, connections):
        queue_root = PIPELINE_TRIGGER_ROOT.format(
            tenant=tenant_name,
            pipeline=pipeline_name)
        super().__init__(client, queue_root, connections)

    @classmethod
    def createRegistry(cls, client, connections):
        """Create a tenant/pipeline queue registry

        Returns a nested dictionary of:
          tenant_name -> pipeline_name -> EventQueue

        Queues are dynamically created with the originally supplied ZK
        client and connection registry as they are accessed via the
        queue registry (so new tenants or pipelines show up
        automatically).

        """
        return DefaultKeyDict(
            lambda t: cls._createRegistry(client, t, connections)
        )

    @classmethod
    def _createRegistry(cls, client, tenant_name, connections):
        return DefaultKeyDict(
            lambda p: cls(client, tenant_name, p, connections)
        )


class ConnectionEventQueue(ZooKeeperEventQueue):
    """Connection events via ZooKeeper"""

    log = logging.getLogger("zuul.ConnectionEventQueue")

    def __init__(self, client, connection_name, component_info):
        queue_root = "/".join((CONNECTION_ROOT, connection_name, "events"))
        super().__init__(client, queue_root)
        self.election_root = "/".join(
            (CONNECTION_ROOT, connection_name, "election")
        )
        self.kazoo_client.ensure_path(self.election_root)
        if component_info:
            self.election = RendezvousElection(
                self.kazoo_client, self.election_root,
                "scheduler", component_info)

    def _eventWatch(self, callback, event_list):
        if event_list:
            return callback()

    def registerEventWatch(self, callback):
        self.kazoo_client.ChildrenWatch(
            self.event_root, functools.partial(self._eventWatch, callback)
        )

    def put(self, data):
        log = self.log
        if "zuul_event_id" in data:
            log = get_annotated_logger(log, data["zuul_event_id"])
        log.debug("Submitting connection event to queue %s: %s",
                  self.event_root, data)
        self._put({'event_data': data})

    def iter(self, event_id_offset=None):
        for data, ack_ref, zstat in self._iterEvents(event_id_offset):
            event = model.ConnectionEvent.fromDict(data['event_data'])
            event.ack_ref = ack_ref
            event.zuul_event_ltime = zstat.creation_transaction_id
            yield event

    def __iter__(self):
        yield from self.iter()


class EventReceiverElection(SessionAwareElection):
    """Election for a singleton event receiver."""

    def __init__(self, client, connection_name, receiver_name):
        self.election_root = "/".join(
            (CONNECTION_ROOT, connection_name, f"election-{receiver_name}")
        )
        super().__init__(client.client, self.election_root)


class NodepoolEventElection(SessionAwareElection):
    """Election for the nodepool completion event processor."""

    def __init__(self, client):
        self.election_root = "/zuul/nodepool/election"
        super().__init__(client.client, self.election_root)


class EventCheckpoint(ZooKeeperSimpleBase):
    """Store checkpoint data for drivers that need it."""

    log = logging.getLogger("zuul.EventCheckpoint")

    def __init__(self, client, connection_name, receiver_name):
        super().__init__(client)
        self.root = "/".join(
            (CONNECTION_ROOT, connection_name, f"checkpoint-{receiver_name}")
        )
        self.stat = None

    def get(self):
        """Return the most recently stored checkpoint value or None"""
        try:
            data, stat = self.kazoo_client.get(self.root)
        except NoNodeError:
            self.stat = None
            return None

        try:
            data = json.loads(data.decode("utf-8"))
        except Exception:
            self.stat = None
            return None

        self.stat = stat
        return data['checkpoint']

    def set(self, checkpoint):
        """Set the checkpoint value

        If it has been updated since this object last read the value,
        Kazoo will raise an exception.

        """

        data = {'checkpoint': checkpoint}
        data = json.dumps(data, sort_keys=True).encode("utf-8")
        version = -1
        if self.stat:
            version = self.stat.version
        try:
            self.stat = self.kazoo_client.set(self.root, data, version)
        except NoNodeError:
            path, self.stat = self.kazoo_client.create(self.root, data,
                                                       include_data=True)



# Copyright 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import contextlib
import os
import time
from unittest import mock

import fixtures

from zuul.driver.openstack import OpenstackDriver
from zuul.driver.openstack.openstackmodel import OpenstackProviderNode

from tests.fake_openstack import (
    FakeOpenstackCloud,
    FakeOpenstackFloatingIp,
    FakeOpenstackPort,
    FakeOpenstackProviderEndpoint,
)
from tests.base import (
    FIXTURE_DIR,
    ResponsesFixture,
    ZuulTestCase,
    driver_config,
    iterate_timeout,
    return_data,
    simple_layout,
)
from tests.unit.test_launcher import ImageMocksFixture
from tests.unit.test_cloud_driver import BaseCloudDriverTest


class BaseOpenstackDriverTest(ZuulTestCase):
    cloud_test_image_format = 'qcow2'
    cloud_test_provider_name = 'openstack-main'
    config_file = 'zuul-connections-nodepool.conf'
    debian_return_data = {
        'zuul': {
            'artifacts': [
                {
                    'name': 'qcow2 image',
                    'url': 'http://example.com/image.qcow2',
                    'metadata': {
                        'type': 'zuul_image',
                        'image_name': 'debian-local',
                        'format': 'qcow2',
                        'sha256': ImageMocksFixture.qcow2_sha256,
                        'md5sum': ImageMocksFixture.qcow2_md5sum,
                    }
                },
            ]
        }
    }
    openstack_needs_floating_ip = False
    openstack_auto_attach_floating_ip = True

    def _makeFakeCloud(self, region):
        fake_cloud = FakeOpenstackCloud(
            region,
            needs_floating_ip=self.openstack_needs_floating_ip,
            auto_attach_floating_ip=self.openstack_auto_attach_floating_ip,
        )
        fake_cloud.max_instances =\
            self.test_config.driver.openstack.get('max_instances', 100)
        fake_cloud.max_cores =\
            self.test_config.driver.openstack.get('max_cores', 100)
        fake_cloud.max_ram =\
            self.test_config.driver.openstack.get('max_ram', 1000000)
        fake_cloud.max_volumes =\
            self.test_config.driver.openstack.get('max_volumes', 100)
        fake_cloud.max_volume_gb =\
            self.test_config.driver.openstack.get('max_volume_gb', 100)
        self.fake_cloud[region] = fake_cloud

    def setUp(self):
        self.initTestConfig()
        responses_fixture = self.useFixture(ResponsesFixture())
        self.useFixture(ImageMocksFixture(responses_fixture))
        clouds_yaml = os.path.join(FIXTURE_DIR, 'clouds.yaml')
        self.useFixture(
            fixtures.EnvironmentVariable('OS_CLIENT_CONFIG_FILE', clouds_yaml))

        self.fake_cloud = {}
        self._makeFakeCloud(None)
        self._makeFakeCloud('region2')
        self.patch(OpenstackDriver, '_endpoint_class',
                   FakeOpenstackProviderEndpoint)
        self.patch(FakeOpenstackProviderEndpoint,
                   '_fake_cloud', self.fake_cloud)
        super().setUp()

    @contextlib.contextmanager
    def _block_futures(self):
        with (mock.patch(
                'zuul.driver.openstack.openstackendpoint.'
                'OpenstackProviderEndpoint._completeApi', return_value=None)):
            yield


class TestOpenstackDriver(BaseOpenstackDriverTest, BaseCloudDriverTest):
    def _assertProviderNodeAttributes(self, pnode):
        super()._assertProviderNodeAttributes(pnode)
        self.assertEqual('fakecloud', pnode.cloud)
        self.assertEqual('region1', pnode.region)
        if checks := self.test_config.driver.openstack.get('node_checks'):
            checks(self, pnode)

    @simple_layout('layouts/openstack/nodepool.yaml', enable_nodepool=True)
    def test_openstack_node_lifecycle(self):
        self._test_node_lifecycle('debian-normal')

    def check_more_attrs(self, pnode):
        self.assertEqual('foo', pnode.az)

    @simple_layout('layouts/openstack/more.yaml', enable_nodepool=True)
    @driver_config('openstack', node_checks=check_more_attrs)
    def test_openstack_node_lifecycle_more(self):
        # Test with more options than the normal test
        self._test_node_lifecycle('debian-normal')

    @simple_layout('layouts/openstack/nodepool.yaml', enable_nodepool=True)
    @driver_config('openstack', max_cores=4)
    def test_openstack_quota(self):
        self._test_quota('debian-normal')

    @simple_layout('layouts/openstack/resource-limits.yaml',
                   enable_nodepool=True)
    def test_openstack_resource_limits(self):
        self._test_quota('debian-normal')

    @simple_layout('layouts/openstack/nodepool-image.yaml',
                   enable_nodepool=True)
    @return_data(
        'build-debian-local-image',
        'refs/heads/master',
        BaseOpenstackDriverTest.debian_return_data,
    )
    def test_openstack_diskimage(self):
        self._test_diskimage()

    # Openstack-driver specific tests

    @simple_layout('layouts/openstack/nodepool-multi.yaml',
                   enable_nodepool=True)
    def test_openstack_resource_cleanup(self):
        self.waitUntilSettled()
        self.launcher.cleanup_worker.INTERVAL = 1
        conn = self.fake_cloud[None]._getConnection()
        # This test class doesn't generally use fips, but since we're
        # testing *all* leaked resources, set it here.
        conn.cloud._fake_needs_floating_ip = True

        system_id = self.launcher.system.system_id
        tags = {
            'zuul_system_id': system_id,
            'zuul_node_uuid': '0000000042',
        }
        flavor = conn.list_flavors()[0]
        conn.create_server(
            name="test",
            meta=tags,
            flavor=flavor,
        )
        self.assertEqual(1, len(conn.list_servers()))

        fip = FakeOpenstackFloatingIp(
            id='42',
            floating_ip_address='fake',
            port_id=None,
            status='ACTIVE',
        )
        conn.cloud.floating_ips.append(fip)
        self.assertEqual(1, len(conn.list_floating_ips()))

        port = FakeOpenstackPort(
            id='43',
            status='DOWN',
            device_owner='compute:foo',
        )
        conn.cloud.ports.append(port)
        self.assertEqual(1, len(conn.list_ports()))

        self.log.debug("Start cleanup worker")
        self.launcher.cleanup_worker.start()

        for _ in iterate_timeout(30, 'instance deletion'):
            if not conn.list_servers():
                break
            time.sleep(1)
        for _ in iterate_timeout(30, 'fip deletion'):
            if not conn.list_floating_ips():
                break
            time.sleep(1)
        for _ in iterate_timeout(30, 'port deletion'):
            if not conn.list_ports():
                break
            time.sleep(1)

    @simple_layout('layouts/openstack/nodepool.yaml', enable_nodepool=True)
    def test_state_machines(self):
        label_name = "debian-normal"
        provider_name = "openstack-main"
        node_class = OpenstackProviderNode
        future_names = ['delete_future', 'create_future']
        self._test_state_machines(label_name, provider_name,
                                  node_class, future_names)


class TestOpenstackDriverFloatingIp(BaseOpenstackDriverTest,
                                    BaseCloudDriverTest):
    # This test is for nova-net clouds with floating ips that require
    # manual attachment.
    openstack_needs_floating_ip = True
    openstack_auto_attach_floating_ip = False

    @simple_layout('layouts/openstack/nodepool.yaml', enable_nodepool=True)
    def test_openstack_fip(self):
        self._test_node_lifecycle('debian-normal')

    @simple_layout('layouts/openstack/nodepool.yaml', enable_nodepool=True)
    def test_state_machines(self):
        label_name = "debian-normal"
        provider_name = "openstack-main"
        node_class = OpenstackProviderNode
        future_names = ['delete_future', 'create_future']
        self._test_state_machines(label_name, provider_name,
                                  node_class, future_names)


class TestOpenstackDriverAutoAttachFloatingIp(BaseOpenstackDriverTest,
                                              BaseCloudDriverTest):
    openstack_needs_floating_ip = True
    openstack_auto_attach_floating_ip = True

    @simple_layout('layouts/openstack/nodepool.yaml', enable_nodepool=True)
    def test_openstack_auto_attach_fip(self):
        self._test_node_lifecycle('debian-normal')

    @simple_layout('layouts/openstack/nodepool.yaml', enable_nodepool=True)
    def test_state_machines(self):
        label_name = "debian-normal"
        provider_name = "openstack-main"
        node_class = OpenstackProviderNode
        future_names = ['delete_future', 'create_future']
        self._test_state_machines(label_name, provider_name,
                                  node_class, future_names)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import types

import zuul.model


class ZuulJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, types.MappingProxyType):
            d = dict(o)
            # Always remove SafeLoader left-over
            d.pop('_source_context', None)
            d.pop('_start_mark', None)
            return d
        elif (
                isinstance(o, zuul.model.SourceContext) or
                isinstance(o, zuul.model.ZuulMark)):
            return {}
        return json.JSONEncoder.default(self, o)


def json_dumps(obj, **kw):
    return json.dumps(obj, cls=ZuulJSONEncoder, **kw)



# Copyright 2024-2025 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import urllib

from zuul.driver import Driver, ConnectionInterface, ProviderInterface
from zuul.driver.static import (
    staticconnection,
    staticmodel,
    staticprovider,
    staticendpoint,
)

from zuul.provider import EndpointCacheMixin


class StaticDriver(Driver, EndpointCacheMixin,
                   ConnectionInterface, ProviderInterface):
    name = 'static'
    _endpoint_class = staticendpoint.StaticProviderEndpoint

    def getConnection(self, name, config):
        return staticconnection.StaticConnection(self, name, config)

    def getProvider(self, zk_client, connection, tenant_name,
                    canonical_name, provider_config, system_id):
        return staticprovider.StaticProvider(
            self, zk_client, connection, tenant_name,
            canonical_name, provider_config, system_id)

    def getProviderClass(self):
        return staticprovider.StaticProvider

    def getProviderSchema(self):
        return staticprovider.StaticProviderSchema().getProviderSchema()

    def getProviderSchemaClass(self):
        return staticprovider.StaticProviderSchema

    def getProviderNodeClass(self):
        return staticmodel.StaticProviderNode

    def getEndpoint(self, provider):
        endpoint_id = urllib.parse.quote_plus(
            provider.connection.connection_name)
        return self.getEndpointById(
            endpoint_id,
            create_args=(self, provider.zk_client, provider.connection,
                         provider.system_id))

    def stop(self):
        self.stopEndpoints()



# Copyright 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
from zuul.lib.config import get_default


def get_statsd_config(config):
    statsd_host = get_default(config, 'statsd', 'server')
    statsd_port = int(get_default(config, 'statsd', 'port', 8125))
    statsd_prefix = get_default(config, 'statsd', 'prefix')
    return (statsd_host, statsd_port, statsd_prefix)


def normalize_statsd_name(name):
    name = name.replace('.', '_')
    name = name.replace(':', '_')
    # Statsd doesn't support non-ascii chars
    name = name.encode('ascii', 'replace').decode()
    return name


def get_statsd(config, extra_keys=None):
    (statsd_host, statsd_port, statsd_prefix) = get_statsd_config(config)
    if statsd_host is None:
        return None
    import statsd

    class CustomStatsClient(statsd.StatsClient):

        def __init__(self, host, port, prefix, extra=None):
            self.extra_keys = copy.copy(extra) or {}

            for key in self.extra_keys:
                value = normalize_statsd_name(self.extra_keys[key])
                self.extra_keys[key] = value

            super().__init__(host, port, prefix)

        def _format_stat(self, name, **keys):
            format_keys = copy.copy(keys)

            # we need to normalize all keys which go into the metric name
            for key in format_keys:
                normalized_value = normalize_statsd_name(format_keys[key])
                format_keys[key] = normalized_value

            format_keys.update(self.extra_keys)
            return name.format(**format_keys)

        def gauge(self, stat, value, rate=1, delta=False, **format_keys):
            stat = self._format_stat(stat, **format_keys)
            super().gauge(stat, value, rate, delta)

        def incr(self, stat, count=1, rate=1, **format_keys):
            stat = self._format_stat(stat, **format_keys)
            super().incr(stat, count, rate)

        def timing(self, stat, delta, rate=1, **format_keys):
            stat = self._format_stat(stat, **format_keys)
            super().timing(stat, delta, rate)

    return CustomStatsClient(
        statsd_host, statsd_port, statsd_prefix, extra_keys)



# Copyright 2016 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from zuul.driver import (
    Driver, ConnectionInterface, SourceInterface, TriggerInterface
)
from zuul.driver.git import gitconnection
from zuul.driver.git import gitmodel
from zuul.driver.git import gitsource
from zuul.driver.git import gittrigger


class GitDriver(Driver, ConnectionInterface, SourceInterface,
                TriggerInterface):
    name = "git"

    def getConnection(self, name, config):
        return gitconnection.GitConnection(self, name, config)

    def getTrigger(self, connection, config=None):
        return gittrigger.GitTrigger(self, connection, config)

    def getSource(self, connection):
        return gitsource.GitSource(self, connection)

    def getTriggerSchema(self):
        return gittrigger.getSchema()

    def getTriggerEventClass(self):
        return gitmodel.GitTriggerEvent

    def getRequireSchema(self):
        return {}

    def getRejectSchema(self):
        return {}



# Copyright 2025 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from uuid import uuid4
import time

from zuul import model
from zuul.launcher.uploadplanner import UploadPlanner
from zuul.provider import (
    BaseImageCopyJob,
    BaseImageImportJob,
    BaseImageUploadJob,
)
from tests.base import BaseTestCase


def get_uuid(job):
    return job.metadata['zuul_upload_uuid']


class Dummy:
    pass


class FakeImageImportJob(BaseImageImportJob):
    def __init__(self, metadata):
        super().__init__()
        self.metadata = metadata


class FakeImageCopyJob(BaseImageCopyJob):
    def __init__(self, metadata):
        super().__init__()
        self.metadata = metadata


class FakeImageUploadJob(BaseImageUploadJob):
    def __init__(self, metadata):
        super().__init__()
        self.metadata = metadata


class FakeProvider:
    def __init__(self, name):
        self.name = name
        image = Dummy()
        image.name = 'image-name'
        image.canonical_name = 'image-canonical-name'
        image.upload_methods = ['copy', 'import', 'upload']
        self.images = {
            image.name: image,
        }
        self.allow_copy = []

    def getImageImportJob(self, url, provider_image, image_name,
                          image_format, metadata, md5, sha256):
        if f'{self.name}-import' in url:
            return FakeImageImportJob(metadata)
        return None

    def getImageCopyJob(self, source_provider, provider_image,
                        image_name, image_format, metadata, md5, sha256):
        # source image id (source_upload.external_id)
        # source region (source_upload.provider.region)
        # new tags
        # don't copy tags
        if source_provider.name in self.allow_copy:
            return FakeImageCopyJob(metadata)
        return None

    def getImageUploadJob(self, provider_image, image_name,
                          image_format, metadata, md5, sha256):
        return FakeImageUploadJob(metadata)


class FakeLauncher:
    def __init__(self):
        self.system = Dummy()
        self.system.system_id = 'system-id'
        self._providers = {
            'p1': FakeProvider('p1'),
            'p2': FakeProvider('p2'),
            'p3': FakeProvider('p3'),
            'p4': FakeProvider('p4'),
        }

    def _getProviderByCanonicalName(self, provider_name):
        return self._providers[provider_name]


class TestUploadPlanner(BaseTestCase):
    def makeArtifact(self, uuid=None, url=None):
        if uuid is None:
            uuid = uuid4().hex
        if url is None:
            url = 'http://example.com/image.raw.zst'
        iba = model.ImageBuildArtifact()
        iba._set(
            uuid=uuid,
            name='image-name',
            canonical_name='image-canonical-name',
            project_canonical_name='project-canonical-name',
            url=url,
            timestamp=time.time(),
        )
        return iba

    def makeUpload(self, iba, config_hash=None, providers=None, uuid=None,
                   state=None, external_id=None):
        if uuid is None:
            uuid = uuid4().hex
        if config_hash is None:
            config_hash = 'config-hash'
        if providers is None:
            providers = ['p1']
        if state is None:
            state = model.ImageUpload.State.PENDING
        upload = model.ImageUpload()
        upload._set(
            uuid=uuid,
            artifact_uuid=iba.uuid,
            endpoint_name='endpoint-name',
            providers=providers,
            canonical_name='image-canonical-name',
            config_hash=config_hash,
            timestamp=time.time(),
            _state=model.ImageUpload.State.PENDING,
            state_time=time.time(),
            external_id=external_id,
        )
        return upload

    def printJobs(self, planner):
        for x in planner.import_jobs:
            self.log.debug("Import job: %s", x[1].metadata)
        for x in planner.upload_jobs:
            self.log.debug("Upload job: %s", x[1].metadata)
        for x in planner.copy_jobs:
            self.log.debug("Copy job: %s", x[1].metadata)

    def test_one_upload(self):
        # There's only one upload, we should have an upload job for it
        iba = self.makeArtifact()
        upload = self.makeUpload(iba)
        launcher = FakeLauncher()
        pending_uploads = [upload]
        all_uploads = [upload]
        planner = UploadPlanner(launcher, iba, pending_uploads, all_uploads)
        planner.plan()

        self.printJobs(planner)
        self.assertEqual([], planner.import_jobs)
        self.assertEqual([upload.uuid],
                         [get_uuid(x[1]) for x in planner.upload_jobs])
        self.assertEqual([], planner.copy_jobs)

    def test_one_import(self):
        # There's only one upload, we should have an import job for it
        iba = self.makeArtifact(url='p1-import')
        upload = self.makeUpload(iba)
        launcher = FakeLauncher()
        pending_uploads = [upload]
        all_uploads = [upload]
        planner = UploadPlanner(launcher, iba, pending_uploads, all_uploads)
        planner.plan()

        self.printJobs(planner)
        self.assertEqual([upload.uuid],
                         [get_uuid(x[1]) for x in planner.import_jobs])
        self.assertEqual([], planner.upload_jobs)
        self.assertEqual([], planner.copy_jobs)

    def test_import_copy(self):
        # We import, then copy from that
        iba = self.makeArtifact(url='p1-import')
        upload_import = self.makeUpload(iba, providers=['p1'])
        upload_copy = self.makeUpload(iba, providers=['p2'])
        launcher = FakeLauncher()
        launcher._providers['p2'].allow_copy.append('p1')
        pending_uploads = [upload_import, upload_copy]
        all_uploads = [upload_import, upload_copy]
        planner = UploadPlanner(launcher, iba, pending_uploads, all_uploads)
        planner.plan()

        self.printJobs(planner)
        self.assertEqual([upload_import.uuid],
                         [get_uuid(x[1]) for x in planner.import_jobs])
        self.assertEqual([], planner.upload_jobs)
        self.assertEqual([], planner.copy_jobs)
        self.assertEqual(
            [upload_copy.uuid],
            [get_uuid(x[1]) for x in planner.import_jobs[0][1].dependents])

    def test_upload_copy(self):
        # We upload, then copy from that
        iba = self.makeArtifact()
        upload_upload = self.makeUpload(iba, providers=['p1'])
        upload_copy = self.makeUpload(iba, providers=['p2'])
        launcher = FakeLauncher()
        launcher._providers['p2'].allow_copy.append('p1')
        pending_uploads = [upload_upload, upload_copy]
        all_uploads = [upload_upload, upload_copy]
        planner = UploadPlanner(launcher, iba, pending_uploads, all_uploads)
        planner.plan()

        self.printJobs(planner)
        self.assertEqual([], planner.import_jobs)
        self.assertEqual([upload_upload.uuid],
                         [get_uuid(x[1]) for x in planner.upload_jobs])
        self.assertEqual([], planner.copy_jobs)
        self.assertEqual(
            [upload_copy.uuid],
            [get_uuid(x[1]) for x in planner.upload_jobs[0][1].dependents])

    def test_multi_import_copy(self):
        # Two imports with copies
        iba = self.makeArtifact(url='p1-import p2-import')
        upload_import1 = self.makeUpload(iba, providers=['p1'])
        upload_import2 = self.makeUpload(iba, providers=['p2'])
        upload_copy1 = self.makeUpload(iba, providers=['p3'])
        upload_copy2 = self.makeUpload(iba, providers=['p4'])
        launcher = FakeLauncher()
        pending_uploads = [upload_import1, upload_copy1,
                           upload_import2, upload_copy2]

        # (from, to)
        launcher._providers['p4'].allow_copy.append('p1')
        launcher._providers['p3'].allow_copy.append('p2')
        all_uploads = pending_uploads[:]
        planner = UploadPlanner(launcher, iba, pending_uploads, all_uploads)
        planner.plan()

        self.printJobs(planner)
        self.assertEqual([upload_import1.uuid, upload_import2.uuid],
                         [get_uuid(x[1]) for x in planner.import_jobs])
        self.assertEqual([], planner.upload_jobs)
        self.assertEqual([], planner.copy_jobs)
        self.assertEqual(
            [upload_copy2.uuid],
            [get_uuid(x[1]) for x in planner.import_jobs[0][1].dependents])
        self.assertEqual(
            [upload_copy1.uuid],
            [get_uuid(x[1]) for x in planner.import_jobs[1][1].dependents])

    def test_multi_upload_copy(self):
        # Two uploads with copies
        iba = self.makeArtifact(url='p1-upload')
        upload_upload1 = self.makeUpload(iba, providers=['p1'])
        upload_upload2 = self.makeUpload(iba, providers=['p2'])
        upload_copy1 = self.makeUpload(iba, providers=['p3'])
        upload_copy2 = self.makeUpload(iba, providers=['p4'])
        launcher = FakeLauncher()
        pending_uploads = [upload_upload1, upload_copy1,
                           upload_upload2, upload_copy2]

        # (from, to)
        launcher._providers['p4'].allow_copy.append('p1')
        launcher._providers['p3'].allow_copy.append('p2')
        all_uploads = pending_uploads[:]
        planner = UploadPlanner(launcher, iba, pending_uploads, all_uploads)
        planner.plan()

        self.printJobs(planner)
        self.assertEqual([upload_upload1.uuid, upload_upload2.uuid],
                         [get_uuid(x[1]) for x in planner.upload_jobs])
        self.assertEqual([], planner.import_jobs)
        self.assertEqual([], planner.copy_jobs)
        self.assertEqual(
            [upload_copy2.uuid],
            [get_uuid(x[1]) for x in planner.upload_jobs[0][1].dependents])
        self.assertEqual(
            [upload_copy1.uuid],
            [get_uuid(x[1]) for x in planner.upload_jobs[1][1].dependents])

    def test_finished_import_copy(self):
        # Copy from a finished import
        iba = self.makeArtifact(url='p1-import')
        upload_import = self.makeUpload(iba, providers=['p1'],
                                        state=model.ImageUpload.State.READY,
                                        external_id='test_external_id')
        upload_copy = self.makeUpload(iba, providers=['p2'])
        launcher = FakeLauncher()
        launcher._providers['p2'].allow_copy.append('p1')
        pending_uploads = [upload_copy]
        all_uploads = [upload_import, upload_copy]
        planner = UploadPlanner(launcher, iba, pending_uploads, all_uploads)
        planner.plan()

        self.printJobs(planner)
        self.assertEqual([],
                         [get_uuid(x[1]) for x in planner.import_jobs])
        self.assertEqual([], planner.upload_jobs)
        self.assertEqual([upload_copy.uuid],
                         [get_uuid(x[1]) for x in planner.copy_jobs])
        for j in planner.copy_jobs:
            self.assertEqual([], j[1].dependents)

    def test_finished_upload_copy(self):
        # Copy from a finished upload
        iba = self.makeArtifact()
        upload_upload = self.makeUpload(iba, providers=['p1'],
                                        state=model.ImageUpload.State.READY,
                                        external_id='test_external_id')
        upload_copy = self.makeUpload(iba, providers=['p2'])
        launcher = FakeLauncher()
        launcher._providers['p2'].allow_copy.append('p1')
        pending_uploads = [upload_copy]
        all_uploads = [upload_upload, upload_copy]
        planner = UploadPlanner(launcher, iba, pending_uploads, all_uploads)
        planner.plan()

        self.printJobs(planner)
        self.assertEqual([],
                         [get_uuid(x[1]) for x in planner.upload_jobs])
        self.assertEqual([], planner.import_jobs)
        self.assertEqual([upload_copy.uuid],
                         [get_uuid(x[1]) for x in planner.copy_jobs])
        for j in planner.copy_jobs:
            self.assertEqual([], j[1].dependents)



# Copyright 2023 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import fixtures
import json
import os
import queue
import time
import threading

import google.api_core.exceptions

import tests.base
from tests.base import (
    ZuulTestCase,
    iterate_timeout,
    okay_tracebacks,
    simple_layout,
)


FIXTURE_DIR = os.path.join(tests.base.FIXTURE_DIR, 'gerrit')


class FakePubsubMessage:
    def __init__(self, data):
        self.data = data
        self._acked = False

    def ack(self):
        self._acked = True


class FakePubsubFuture:
    def __init__(self):
        self.event = threading.Event()
        self._error = None

    def result(self):
        self.event.wait()
        if self._error:
            raise self._error

    def cancel(self, error=None):
        self._error = error
        self.event.set()


class FakePubsubSubscriber:
    def __init__(self, credentials=None):
        self.credentials = credentials
        self._queue = queue.Queue()
        self._closed = False
        self._entered = False

    def put(self, data):
        self._queue.put(data)

    def create_subscription(self, name=None, topic=None):
        if not self._entered:
            raise Exception("Attempt to use subscriber "
                            "outside of context manager")
        self._name = name
        self._topic = topic
        raise google.api_core.exceptions.AlreadyExists("Exists")

    def subscribe(self, name, callback):
        if self._closed:
            raise Exception("Attempt to use closed subscriber")
        if not self._entered:
            raise Exception("Attempt to use subscriber "
                            "outside of context manager")
        assert self._name == name
        self._callback = callback
        self._future = FakePubsubFuture()
        self._thread = threading.Thread(target=self._run)
        self._thread.start()
        return self._future

    def _run(self):
        while not self._closed:
            data = self._queue.get()
            self._queue.task_done()
            if data is None:
                continue
            msg = FakePubsubMessage(data)
            self._callback(msg)

    def __enter__(self):
        self._entered = True
        return self

    def __exit__(self, *args, **kw):
        self._closed = True
        self._queue.put(None)


def serialize(event):
    return json.dumps(event)


class TestGerritEventSourceGcloudPubsub(ZuulTestCase):
    config_file = 'zuul-gerrit-gcloud.conf'

    def setUp(self):
        self.useFixture(fixtures.MonkeyPatch(
            'zuul.driver.gerrit.gerriteventgcloudpubsub.'
            'pubsub_v1.SubscriberClient',
            FakePubsubSubscriber))
        super().setUp()

    @simple_layout('layouts/simple.yaml')
    @okay_tracebacks('Test error')
    def test_gcloud_pubsub(self):
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A')
        subscriber = self.fake_gerrit.event_thread.subscriber
        self.fake_gerrit.event_thread.RECONNECTION_DELAY = 1

        # Assert we passed the required config entries
        self.assertEqual('projects/testproject/subscriptions/zuul',
                         subscriber._name)
        self.assertEqual('projects/testproject/topics/gerrit',
                         subscriber._topic)

        # Exercise reconnection
        err = Exception("Test error")
        subscriber._future.cancel(err)

        for _ in iterate_timeout(60, 'wait for reconnect'):
            if subscriber is not self.fake_gerrit.event_thread.subscriber:
                break
            time.sleep(0.2)

        subscriber = self.fake_gerrit.event_thread.subscriber
        self.additional_event_queues.append(subscriber._queue)

        subscriber.put(serialize(A.getPatchsetCreatedEvent(1)))
        self.waitUntilSettled()

        self.assertHistory([
            dict(name='check-job', result='SUCCESS', changes='1,1')
        ])
        self.assertEqual(A.reported, 1, "A should be reported")

        self.assertTrue(subscriber._queue.empty())



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import daemon
import logging
import os
import sys

import extras
import fixtures
import testtools

from tests.base import iterate_timeout

# as of python-daemon 1.6 it doesn't bundle pidlockfile anymore
# instead it depends on lockfile-0.9.1 which uses pidfile.
pid_file_module = extras.try_imports(['daemon.pidlockfile', 'daemon.pidfile'])


def daemon_test(pidfile, flagfile):
    pid = pid_file_module.TimeoutPIDLockFile(pidfile, 10)
    with daemon.DaemonContext(pidfile=pid):
        for x in iterate_timeout(30, "flagfile to be removed"):
            if not os.path.exists(flagfile):
                break
    sys.exit(0)


class TestDaemon(testtools.TestCase):
    log = logging.getLogger("zuul.test.daemon")

    def setUp(self):
        super(TestDaemon, self).setUp()
        self.test_root = self.useFixture(fixtures.TempDir(
            rootdir=os.environ.get("ZUUL_TEST_ROOT"))).path

    def test_daemon(self):
        pidfile = os.path.join(self.test_root, "daemon.pid")
        flagfile = os.path.join(self.test_root, "daemon.flag")
        open(flagfile, 'w').close()
        if not os.fork():
            self._cleanups = []
            daemon_test(pidfile, flagfile)
        for x in iterate_timeout(30, "daemon to start"):
            if os.path.exists(pidfile):
                break
        os.unlink(flagfile)
        for x in iterate_timeout(30, "daemon to stop"):
            if not os.path.exists(pidfile):
                break



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Based on ASL2 code from:
# https://gist.github.com/tim-patterson/4471877

import sys
import io
import threading
import socketserver
import code


class ThreadLocalProxy(object):
    def __init__(self, default):
        self.files = {}
        self.default = default

    def __getattr__(self, name):
        obj = self.files.get(threading.current_thread(), self.default)
        return getattr(obj, name)

    def register(self, obj):
        self.files[threading.current_thread()] = obj

    def unregister(self):
        self.files.pop(threading.current_thread())


class REPLHandler(socketserver.StreamRequestHandler):
    def handle(self):
        sys.stdout.register(io.TextIOWrapper(self.wfile, 'utf8'))
        sys.stderr.register(io.TextIOWrapper(self.wfile, 'utf8'))
        sys.stdin.register(io.TextIOWrapper(self.rfile, 'utf8'))
        try:
            console = code.InteractiveConsole(locals())
            console.interact('Console:')
        except Exception:
            pass
        finally:
            sys.stdout.unregister()
            sys.stderr.unregister()
            sys.stdin.unregister()


class REPLThreadedTCPServer(socketserver.ThreadingMixIn,
                            socketserver.TCPServer):
    daemon_threads = True
    allow_reuse_address = True

    def __init__(self, scheduler, *args, **kw):
        self.scheduler = scheduler
        super(REPLThreadedTCPServer, self).__init__(*args, **kw)
        sys.stdout = ThreadLocalProxy(sys.stdout)
        sys.stderr = ThreadLocalProxy(sys.stderr)
        sys.stdin = ThreadLocalProxy(sys.stdin)


class REPLServer(object):
    def __init__(self, scheduler):
        self.server = REPLThreadedTCPServer(
            scheduler, ('localhost', 3000), REPLHandler)

    def start(self):
        self.thread = threading.Thread(target=self.server.serve_forever)
        self.thread.daemon = True
        self.thread.start()

    def stop(self):
        self.server.shutdown()
        self.server.server_close()
        self.thread.join(10)
